import { world } from "@minecraft/server"
import { wanderingTrades, construction, items, equipments, nature, players } from "./data"


const dimension = world.getDimension("overworld")

function getWorldDayTime() {
	const daytime = world.getTimeOfDay() + 6000
	const datetime = new Date(daytime * 3.6 * 1000)
	const hours = datetime.getHours() < 10 ? "0" + datetime.getHours() : datetime.getHours()
	const minutes = datetime.getMinutes() < 10 ? "0" + datetime.getMinutes() : datetime.getMinutes()
	const day = world.getDay()
	const time = `§7Day: ${day} / ${hours}:${minutes}`
	return time
}

function sendNotification(app, p1Id, p2Id) {
	const players = world.getPlayers()
	const playersId = players.map(p => p.id)
	const p2 = playersId.findIndex(id => id === p2Id)
	if (p2 === -1) return

	const hasNotifications = players[p2].getDynamicProperty(`notification:${app}`) ?? true
	const muted = Conversations.muteStatus(p2Id, p1Id) ?? false
	const message = app === "messages" ? "§6[SMARTPHONE] §rYou have a new message!" : (app === "bank" ? "§6[SMARTPHONE] §rYou have received money!" : "§6[SMARTPHONE] §rYour item has been sold!")
	let hasSmartphone = false
	const container = players[p2].getComponent("inventory").container
	for (let i = 0; i < container.size; i++) {
		if (container.getItem(i)?.typeId === "coptaine:smartphone") {
			hasSmartphone = true
			break
		}
	}
	if (((app === "messages" && !muted) || app === "bank" || app === "store") && hasNotifications && hasSmartphone) {
		players[p2].sendMessage(`${message}`)
		players[p2].playSound("random.pop", { volume: 1.0, pitch: 0.85 })
	}
}

function toRoman(num) {
	if (num === 1) return "I"
	if (num === 2) return "II"
	if (num === 3) return "III"
	if (num === 4) return "IV"
	if (num === 5) return "V"
}

function transformSnakeCase(str) {
	const words = str.split('_')
	const capitalizedWords = words.map(word =>
		word.charAt(0).toUpperCase() + word.slice(1).toLowerCase()
	)
	return capitalizedWords.join(' ')
}

function getShopStorage() {
	const shopStorage = dimension.getEntities({ type: "coptaine:shop", tags: ["storage"] })
	return shopStorage
}


export class PlayersRegistry {
	static register(id, name) {
		world.setDynamicProperty(`registry:${id}`, name)
	}

	static checkStatus(id) {
		const status = this.getRegistered().ids.includes(id)
		return status
	}

	static getRegistered() {
		const registeredPlayers = world.getDynamicPropertyIds().filter(i => i.startsWith("registry:"))
		const names = registeredPlayers.map(n => world.getDynamicProperty(n))
		const ids = registeredPlayers.map(i => i.replace("registry:", ""))
		return { names, ids }
	}
}

export class Conversations {
	static getMyConversations(id) {
		const myConversations = world.getDynamicPropertyIds().filter(c => c.startsWith(`conversation:${id}`))
		const rawConversations = myConversations.map(conversation => ({
			conversation,
			date: Conversations.getMessage(id, conversation.replace(`conversation:${id}_`, ""))["date"]
		}))
		rawConversations.sort((a, b) => b.date - a.date)
		const sortedConversations = rawConversations.map(c => c.conversation)
		return sortedConversations
	}

	static sendMessage(p1Name, p2Name, p1Id, p2Id, newMessage) {
		function updateConversation(message, newMessage) {
			const messages = message.split("\n§l")
			const MAX_MESSAGES = 5000
			messages.push(`${newMessage}`)
			if (messages.length > MAX_MESSAGES || messages[0] === "") messages.shift()
			return messages.join("\n")
		}
		const initialData = {
			"senderName": "",
			"receiverName": "",
			"senderId": "",
			"receiverId": "",
			"message": "",
			"newMessage": "",
			"date": 0
		}
		const prevMessages1 = world.getDynamicProperty(`conversation:${p1Id}_${p2Id}`) ?? JSON.stringify(initialData)
		const prevMessages2 = world.getDynamicProperty(`conversation:${p2Id}_${p1Id}`) ?? JSON.stringify(initialData)
		const newMessages1 = updateConversation(JSON.parse(prevMessages1)["message"], `§r§l[${p1Name}] ${getWorldDayTime()}§r\n${newMessage}`)
		const newMessages2 = updateConversation(JSON.parse(prevMessages2)["message"], `§r§l[${p1Name}] ${getWorldDayTime()}§r\n${newMessage}`)
		function createMessageData(senderName, senderId, receiverName, receiverId, message, newMessage) {
			return {
				senderName,
				receiverName,
				senderId,
				receiverId,
				message,
				newMessage,
				date: Date.now()
			};
		}
		const data1 = createMessageData(p1Name, p1Id, p2Name, p2Id, newMessages1, newMessage);
		const data2 = createMessageData(p2Name, p2Id, p1Name, p1Id, newMessages2, newMessage);
		world.setDynamicProperty(`conversation:${p1Id}_${p2Id}`, JSON.stringify(data1))
		world.setDynamicProperty(`conversation:${p2Id}_${p1Id}`, JSON.stringify(data2))
		sendNotification("messages", p1Id, p2Id)
	}

	static getMessage(p1Id, p2Id) {
		const data = JSON.parse(world.getDynamicProperty(`conversation:${p1Id}_${p2Id}`))
		return data
	}

	static muteStatus(p1Id, p2Id) {
		const isMuted = world.getDynamicProperty(`mute:${p1Id}_${p2Id}`) ?? false
		return isMuted
	}

	static muteConversation(p1Id, p2Id) {
		world.setDynamicProperty(`mute:${p1Id}_${p2Id}`, true)
	}

	static unmuteConversation(p1Id, p2Id) {
		world.setDynamicProperty(`mute:${p1Id}_${p2Id}`, false)
	}

	static blockStatus(p1Id, p2Id) {
		const isBlocked = world.getDynamicProperty(`block:${p1Id}_${p2Id}`) ?? false
		return isBlocked
	}

	static blockConversation(p1Id, p2Id) {
		world.setDynamicProperty(`block:${p1Id}_${p2Id}`, true)
	}

	static unblockConversation(p1Id, p2Id) {
		world.setDynamicProperty(`block:${p1Id}_${p2Id}`, false)
	}
}

export class Notes {
	static addNote(id, title, message) {
		const myNotes = this.getMyNotes(id)
		const newNote = {
			"title": title,
			"message": message,
		}
		myNotes.unshift(newNote)
		this.serializeData(id, myNotes)
	}

	static editNote(id, order, title, message) {
		const myNotes = this.getMyNotes(id)
		const updatedNote = {
			"title": title,
			"message": message,
		}
		myNotes.splice(order, 1)
		myNotes.unshift(updatedNote)
		this.serializeData(id, myNotes)
	}

	static removeNote(id, order) {
		const myNotes = this.getMyNotes(id)
		myNotes.splice(order, 1)
		this.serializeData(id, myNotes)
	}

	static getMyNotes(id) {
		const data = world.getDynamicProperty(`notes:${id}`)
		if (data) return JSON.parse(data).notes
		return []
	}

	static serializeData(id, data) {
		const newData = {
			"notes": data
		}
		const dynamicData = JSON.stringify(newData)
		world.setDynamicProperty(`notes:${id}`, dynamicData)
	}
}

export class Camera {
	static addCamera(ownerId, id, location, rotation, dimension, tpLocation) {
		const myCameras = this.getMyCameras(ownerId)
		const newCamera = {
			"name": `New Camera ${myCameras.length + 1}`,
			"id": id,
			"tpLocation": tpLocation,
			"location": { "x": Math.floor(location.x), "y": Math.floor(location.y), "z": Math.floor(location.z) },
			"rotation": { "x": rotation.x, "y": rotation.y },
			"dimension": dimension,
			"isVisible": true
		}
		myCameras.push(newCamera)
		this.serializeData(ownerId, myCameras)
	}

	static editCamera(id, order, property, value) {
		const myCameras = this.getMyCameras(id)
		myCameras[order][property] = value
		this.serializeData(id, myCameras)
	}

	static removeCamera(id, order) {
		const myCameras = this.getMyCameras(id)
		myCameras.splice(order, 1)
		this.serializeData(id, myCameras)
	}

	static getMyCameras(id) {
		const data = world.getDynamicProperty(`cameras:${id}`)
		if (data) return JSON.parse(data).cameras
		return []
	}

	static serializeData(id, data) {
		const newData = {
			"cameras": data
		}
		const dynamicData = JSON.stringify(newData)
		world.setDynamicProperty(`cameras:${id}`, dynamicData)
	}
}

export class Bank {
	static balance(id) {
		const amount = world.getDynamicProperty(`bank:${id}`) ?? 0
		return amount
	}

	static transfer(amount, p1Id, p2Id, p1Name, p2Name) {
		function updateTransaction(transaction, newTransaction) {
			const transactions = transaction.split("\n")
			const MAX_TRANSACTIONS = 50
			transactions.push(`${newTransaction}`)
			if (transactions.length > MAX_TRANSACTIONS || transactions[0] === "") transactions.shift()
			return transactions.join("\n")
		}
		function process(player, name, fixedAmount, action, des) {
			const balance = world.getDynamicProperty(`bank:${player}`) ?? 0
			const prevTransactions = world.getDynamicProperty(`bank_transactions:${player}`) ?? ""
			const transactions = updateTransaction(prevTransactions, `${getWorldDayTime()}§r${action}${amount}${des}${name}`)
			world.setDynamicProperty(`bank:${player}`, balance + fixedAmount)
			world.setDynamicProperty(`bank_transactions:${player}`, `${transactions}`)
		}
		process(p1Id, p2Name, -amount, "§r Sent §c", "§r to ")
		process(p2Id, p1Name, amount, "§r Received §a", "§r from ")
		sendNotification("bank", p1Id, p2Id)
	}

	static transactions(id) {
		const result = world.getDynamicProperty(`bank_transactions:${id}`)
		return result
	}

	static openVault(player) {
		const vaults = dimension.getEntities({ type: "coptaine:vault", tags: [`tag:${player.id}`] })
		const y = player.getRotation().y
		const location = player.location
		if (!vaults.length) {
			const vault = dimension.spawnEntity("coptaine:vault", { x: Math.floor(location.x), y: Math.floor(location.y), z: Math.floor(location.z)})
			vault.nameTag = `${player.name}'s Vault`
			vault.addTag(`tag:${player.id}`)
		} else {
			vaults[0].nameTag = `${player.name}'s Vault`
			vaults[0].teleport({ x: Math.floor(location.x), y: Math.floor(location.y), z: Math.floor(location.z)}, { rotation: { x: 0, y: Math.round(y / 90) * 90}})
		}
	}
}

export class Marketplace {
	static getData(source, marketName) {
		const data = source
		const ids = data.map(x => x.id)
		const names = data.map(x => x.name)
		const amounts = data.map(x => `x${x.amount}`)
		const prices = data.map(x => `§g$${x.price}`)
		const traderName = marketName === "Wandering Trader" ? "§nWandering Trader" : "§nTrader"
		return { ids, names, amounts, prices, marketName, traderName }
	}

	static wanderingTrades() {
		return wanderingTrades
	}

	static construction() {
		return construction
	}

	static equipment() {
		return equipments
	}

	static items() {
		return items
	}

	static nature() {
		return nature
	}

	static player() {
		return players
	}

	static getMyShop(id) {
		const shop = world.getDynamicPropertyIds().filter(n => n.startsWith(`shop:${id}`))
		return shop
	}

	static editMyShop(shopData, itemData, id) {
		const myItems = this.getShopData(`shop:${id}`)?.items ?? []
		if (Array.isArray(itemData)) {
			const getId = itemData[0]
			const getNameTag = itemData[1] ?? "None"
			const getEnchantments = itemData[2]?.length ? itemData[2] : []
			const getLores = itemData[3]?.length ? itemData[3] : []
			const getPrice = itemData[4]
			const getAmount = itemData[5]
			const newItemData = {
				"id": getId,
				"enchantments": getEnchantments ? getEnchantments.map(e => `${transformSnakeCase(e.type.id)} ${toRoman(e.level)}`) : "",
				"nameTag": getNameTag,
				"lores": getLores,
				"price": getPrice,
				"amount": getAmount,
				"slot": getShopStorage()[0].getDynamicProperty("first_empty_slot")
			}
			myItems.push(newItemData)
		}
		if (typeof itemData === "number") {
			myItems.splice(itemData, 1)
		}
		const newShopData = {
			"name": shopData[0],
			"message": shopData[1],
			"id": id,
			"owner": shopData[2],
			"date": Date.now(),
			"items": myItems
		}
		const dynamicData = JSON.stringify(newShopData)
		world.setDynamicProperty(`shop:${id}`, dynamicData)
	}

	static addItem(player, item, playerSlot) {
		let shop = getShopStorage()[0]
		if (!getShopStorage().length) {
			shop = dimension.spawnEntity("coptaine:shop", player.location)
			shop.addTag("storage")
		}
		const storage = shop.getComponent("inventory").container
		function getFirstEmptySlot() {
			for (let i = 0; i < storage.size; i++) {
				if (!storage.getItem(i)) return i
			}
		}
		shop.setDynamicProperty("first_empty_slot", getFirstEmptySlot())
		storage.setItem(getFirstEmptySlot(), item)
		player.getComponent("inventory").container.setItem(playerSlot, null)
	}

	static getItemInfo(data) {
		const { id, nameTag, enchantments, lores, amount, price } = data
		const rawMessage = {
			rawtext: [
				{ text: `§7ID: §a${id}` },
				{ text: `\n§7Name Tag: §r${nameTag}` },
				{ text: `\n§7Amount: §r${amount}` },
				{ text: `\n§7Price: §g$${price}` }
			]
		}
		if (enchantments.length) {
			rawMessage.rawtext.push({ text: `\n\n§7Enchantments: ` })
			for (const enchantment of enchantments) {
				rawMessage.rawtext.push({ text: `\n§5${enchantment}` })
			}
		}
		if (lores.length) {
			rawMessage.rawtext.push({ text: `\n\n§7Lore: ` })
			for (const lore of lores) {
				rawMessage.rawtext.push({ text: `\n§b${lore}` })
			}
		}
		return rawMessage
	}

	static retrieveItem(playerStorage, slot, sold, id = undefined) {
		const shopStorage = world.getDimension("overworld").getEntities({ type: "coptaine:shop", tags: ["storage"] })
		const storage = shopStorage[0].getComponent("inventory").container
		const item = storage.getItem(slot)
		playerStorage.addItem(item)
		storage.setItem(slot, null)
		if (sold) sendNotification("store", 0, id)
	}

	static getShops() {
		const shops = world.getDynamicPropertyIds().filter(s => s.startsWith("shop:") && this.getShopData(s).items.length)
		const rawShops = shops.map(shop => ({
			shop,
			date: this.getShopData(shop).date
		}))
		rawShops.sort((a, b) => b.date - a.date)
		const sortedShops = rawShops.map(s => s.shop)
		return sortedShops
	}

	static getShopData(shop) {
		const data = world.getDynamicProperty(shop)
		if (data) return JSON.parse(data)
		return undefined
	}

	static getShopStats(shop) {
		const data = world.getDynamicProperty(shop)
		if (data) return JSON.parse(data)
		const newData = {
			"visits": 0,
			"sold": 0,
			"totalSales": 0
		}
		return newData
	}

	static updateShopStats(shop, stat, price = 0) {
		const data = this.getShopStats(shop)
		const newShopStats = {
			"visits": stat === "visits" ? data.visits + 1 : data.visits,
			"sold": stat === "sold" ? data.sold + 1 : data.sold,
			"totalSales": stat === "totalSales" ? Number(data.totalSales + price) : data.totalSales
		}
		const dynamicData = JSON.stringify(newShopStats)
		world.setDynamicProperty(shop, dynamicData)
	}
}

export class Clock {
	static addAlarm(id, name, time, sound, isRepeating, isActive, isRinging) {
		const myAlarms = this.getMyAlarms(id)
		const newAlarm = {
			"name": name,
			"time": `${time[0]}:${time[1]}`,
			"sound": { "name": sound[0], "value": sound[1] },
			"isRepeating": isRepeating,
			"isActive": isActive,
			"isRinging": isRinging
		}
		myAlarms.push(newAlarm)
		this.serializeData(id, myAlarms)
	}

	static editAlarm(id, order, name, time, sound, isRepeating, isActive, isRinging) {
		this.removeAlarm(id, order)
		this.addAlarm(id, name, time, sound, isRepeating, isActive, isRinging)
	}

	static removeAlarm(id, order) {
		const myAlarms = this.getMyAlarms(id)
		myAlarms.splice(order, 1)
		this.serializeData(id, myAlarms)
	}

	static getMyAlarms(id) {
		const data = world.getDynamicProperty(`alarms:${id}`)
		if (data) return JSON.parse(data).alarms
		return []
	}

	static serializeData(id, data) {
		const newData = {
			"alarms": data
		}
		const dynamicData = JSON.stringify(newData)
		world.setDynamicProperty(`alarms:${id}`, dynamicData)
	}
}



