import { world, system, ItemStack } from "@minecraft/server"
import { ActionFormData, ModalFormData } from "@minecraft/server-ui"
import { PlayersRegistry, Conversations, Bank, Marketplace, Notes, Camera, Clock } from "./classes"


const PostWorld = world.afterEvents


const vanillaFoods = [
    "minecraft:apple",
    "minecraft:enchanted_golden_apple",
    "minecraft:baked_potato",
    "minecraft:beef",
    "minecraft:beetroot",
    "minecraft:beetroot_seeds",
    "minecraft:beetroot_soup",
    "minecraft:bread",
    "minecraft:carrot",
    "minecraft:chicken",
    "minecraft:chorus_fruit",
    "minecraft:tropical_fish",
    "minecraft:cooked_beef",
    "minecraft:cooked_chicken",
    "minecraft:cooked_fish",
    "minecraft:cooked_porkchop",
    "minecraft:cooked_rabbit",
    "minecraft:cooked_salmon",
    "minecraft:cookie",
    "minecraft:dried_kelp",
    "minecraft:cod",
    "minecraft:glow_berries",
    "minecraft:golden_apple",
    "minecraft:golden_carrot",
    "minecraft:honey_bottle",
    "minecraft:honeycomb",
    "minecraft:melon",
    "minecraft:melon_seeds",
    "minecraft:mushroom_stew",
    "minecraft:cooked_mutton",
    "minecraft:mutton",
    "minecraft:nether_wart",
    "minecraft:pitcher_pod",
    "minecraft:poisonous_potato",
    "minecraft:porkchop",
    "minecraft:potato",
    "minecraft:pufferfish",
    "minecraft:pumpkin_pie",
    "minecraft:pumpkin_seeds",
    "minecraft:rabbit",
    "minecraft:rabbit_stew",
    "minecraft:rotten_flesh",
    "minecraft:salmon",
    "minecraft:spider_eye",
    "minecraft:suspicious_stew",
    "minecraft:sweet_berries",
    "minecraft:torchflower_seeds",
    "minecraft:wheat_seeds"
]

const alarmSounds = [
    { name: "Alarm 1", value: "coptaine_us.alarm.alarm1" },
    { name: "Alarm 2", value: "coptaine_us.alarm.alarm2" },
    { name: "Alarm 3", value: "coptaine_us.alarm.alarm3" }
]

const musicTracks = [
    "13",
    "cat",
    "block",
    "chirp",
    "far",
    "mall",
    "mellohi",
    "stal",
    "strad",
    "ward",
    "11",
    "wait",
    "otherside",
    "5",
    "pigstep",
    "relic"
]

const moonPhases = [
    "Full Moon",
    "Waning Gibbous",
    "First Quarter",
    "Waning Crescent",
    "New Moon",
    "Waxing Crescent",
    "Last Quarter",
    "Waxing Gibbous",
]
// Random number in between
function rand(min, max) {
    return Math.floor(Math.random() * (max - min + 1)) + min
}
// Clamp number
function clamp(num, min, max) {
    return Math.min(Math.max(num, min), max)
}
// Check if textbox is empty
function isNonEmptyString(value) {
    return typeof value === "string" && value.trim() !== ""
}
// Check if string has letter
function hasLetters(str) {
    return /[a-zA-Z]/.test(str)
}
// Check if string has numbers only
function isNumeric(str) {
    const regex = /^\d+$/
    return regex.test(str)
}
// Get random items in the array
function getRandomItems(arr, quantity) {
    for (let i = arr.length - 1; i > 0; i--) {
        const j = Math.floor(Math.random() * (i + 1))
        [arr[i], arr[j]] = [arr[j], arr[i]]
    }
    return arr.slice(0, quantity)
}
// Combine arrays
function combineArrays(arr1, arr2, arr3) {
    const combinedArray = []
    for (let i = 0; i < arr1.length; i++) {
        combinedArray.push(`${arr1[i]} ${arr2[i]} ${arr3[i]}`)
    }
    return combinedArray
}
// Get timestamp
function getWorldDayTime() {
    const daytime = world.getTimeOfDay() + 6000
    const datetime = new Date(daytime * 3.6 * 1000)
    const hours = datetime.getHours() < 10 ? "0" + datetime.getHours() : datetime.getHours()
    const minutes = datetime.getMinutes() < 10 ? "0" + datetime.getMinutes() : datetime.getMinutes()
    const day = world.getDay()
    const time = `§7Day: §r${day} / ${hours}:${minutes}`
    return time
}
// Get top-most solid block above the player
function getTopMostSolidBlock(dimension, x, z) {
    for (let i = 320; i > -65; i--) {
        const block = dimension.getBlock({ x: x, y: i, z: z })
        if (i == -64 || !block.isAir) return i + 1
    }
}
// Get spawnpoint with dimension
function getSpawnPointAndDimension(getSpawnPoint) {
    const { x, y, z, dimension: spawnDim } = getSpawnPoint
    const spawnDimLoc = spawnDim.id === "minecraft:unset" ? `Unset` : `§4${x}, §2${y}, §9${z} §r(${getDimensionName(spawnDim)}§r)`
    const spawnPoint = `§7Spawnpoint: §r${spawnDimLoc}`
    return spawnPoint
}
// Get dimension name
function getDimensionName(dimension) {
    const dimensionId = dimension.id
    const color = dimensionId === "minecraft:overworld" ? "§a" : (dimensionId === "minecraft:the_end" ? "§5" : "§4")
    const removeMc = dimensionId.slice(10).replace("_", " ").replace(" e", " E")
    const name = removeMc[0].toUpperCase() + removeMc.slice(1)
    return `${color}${name}`
}
// Moon Phase
function getMoonPhase() {
    return moonPhases[world.getMoonPhase()]
}
// Get player dynamics
function getPlayerStats(player) {
    const getSpawnPoint = player.getSpawnPoint() ?? {
        x: "unset", y: "unset", z: "unset", dimension: { id: "minecraft:unset" }
    }
    const getWeather = player.getDynamicProperty("coptaine:weather")
    const getLight = player.getDynamicProperty("coptaine:light")
    const getBiome = player.getDynamicProperty("coptaine:biome")
    const getTemperature = player.getDynamicProperty("coptaine:temperature")
    return { getSpawnPoint, getWeather, getLight, getBiome, getTemperature }
}
// Get weather at position
function weatherAtPosition(player) {
    const { location, dimension } = player
    const { x, y, z } = location
    const topBlock = dimension.id !== "overworld" ? y : getTopMostSolidBlock(dimension, x, z)
    const spawnSensor = dimension.spawnEntity("coptaine:sensor", { x: x, y: y, z: z })
    system.runTimeout(() => {
        spawnSensor.teleport({ x: x, y: topBlock, z: z })
        spawnSensor.nameTag = String(y)
    }, 1)
}
// Extract ID
function extractId(conversation, p1Id) {
    const id = conversation.replace(`conversation:${p1Id}_`, "")
    return id
}
// Get time played
function getTimePlayed(player) {
    const timePlayed = player.getDynamicProperty("stats:time_played") ?? 0
    player.setDynamicProperty("stats:time_played", timePlayed + 1)
}
// Get time since last death
function getTimeSinceLastDeath(player) {
    const timePlayed = player.getDynamicProperty("stats:time_since_last_death") ?? 0
    player.setDynamicProperty("stats:time_since_last_death", timePlayed + 1)
}
// Format note message for printing
function splitString(input, max = 30) {
    let result = "";
    let currentLine = "";
    let wordBuffer = "";
    let specialChar = "";
    for (let i = 0; i < input.length; i++) {
        if (input[i] === "|") {
            currentLine += wordBuffer;
            wordBuffer = "";
            result += currentLine + "\n";
            currentLine = "";
            continue;
        }
        if (input[i] === " " || input[i] === "\n") {
            currentLine += wordBuffer + input[i];
            wordBuffer = "";
        } else {
            wordBuffer += input[i];
        }
        if (input[i] === "§" && i + 1 < input.length && /^[a-zA-Z0-9]$/.test(input[i + 1])) {
            specialChar = input[i] + input[i + 1];
        }
        if (currentLine.length + wordBuffer.length > max) {
            if (currentLine.length > 0) {
                result += currentLine + "\n";
                currentLine = "";
            } else {
                result += wordBuffer + "\n";
                wordBuffer = "";
            }
            if (specialChar) {
                currentLine += specialChar;
            }
        }
    }
    result += currentLine + wordBuffer;
    return result;
}
// Block face
function getFaceHitBlock(player) {
    const block = player.getBlockFromViewDirection(
        {
            includeLiquidBlocks: false,
            includePassableBlocks: false,
            maxDistance: 7
        }
    )
    const face = block?.face
    const b = block?.block
    if (!block) return
    let location = ""
    let orientation = ""
    if (face == "North") {
        location = { x: b.x + 0.5, y: b.y, z: b.z - 0.5 }
        orientation = "e:north"
        return { location, orientation }
    }
    if (face == "South") {
        location = { x: b.x + 0.5, y: b.y, z: b.z + 1.5 }
        orientation = "e:south"
        return { location, orientation }
    }
    if (face == "East") {
        location = { x: b.x + 1.5, y: b.y, z: b.z + 0.5 }
        orientation = "e:east"
        return { location, orientation }
    }
    if (face == "West") {
        location = { x: b.x - 0.5, y: b.y, z: b.z + 0.5 }
        orientation = "e:west"
        return { location, orientation }
    }
    if (face == "Up") {
        location = { x: b.x + 0.5, y: b.y + 1, z: b.z + 0.5 }
        orientation = "e:up"
        return { location, orientation }
    }
    if (face == "Down") {
        location = { x: b.x + 0.5, y: b.y, z: b.z + 0.5 }
        orientation = "e:down"
        return { location, orientation }
    }
}
// Get innventory
function getInventory(entity) {
    const container = entity.getComponent("inventory").container
    const items = []
    const itemSlots = []
    for (let i = 0; i < container.size; i++) {
        const item = container.getItem(i)
        if (item) {
            items.push(item)
            itemSlots.push(i)
        }
    }
    return { container, items, itemSlots }
}
// Deduct item in hand
function deductItemInHand(player) {
    if (player.matches({ gameMode: "creative" })) return
    const equip = player.getComponent("equippable")
    const getItem = equip.getEquipment("Mainhand")
    const amount = getItem?.amount
    if (amount > 1) getItem.amount--
    equip.setEquipment("Mainhand", amount > 1 ? getItem : null)
}


function smartphone(player, shortcut = undefined) {
    const { id: userId, name: userName, dimension, location } = player
    if (shortcut) {
        if (shortcut === "camera") return apps("camera")
    }
    const registered = PlayersRegistry.checkStatus(userId)
    if (!registered) return intro()
    const { getSpawnPoint, getWeather, getLight, getBiome, getTemperature } = getPlayerStats(player)
    const locationWidget = player.getDynamicProperty("widget:location")
    const dimensionWidget = player.getDynamicProperty("widget:dimension")
    const biomerWidget = player.getDynamicProperty("widget:biome")
    const weatherWidget = player.getDynamicProperty("widget:weather")
    const lightWidget = player.getDynamicProperty("widget:light")
    const moontWidget = player.getDynamicProperty("widget:moon_phase")
    const spawnpointWidget = player.getDynamicProperty("widget:spawnpoint")
    const timeStamp = getWorldDayTime()
    const owner = `§7Owner: §r${userName}`
    const loc = `§7Location: §4${Math.floor(location.x)}, §2${Math.floor(location.y)}, §9${Math.floor(location.z)}`
    const dim = `§7Dimension: ${getDimensionName(dimension)}`
    const biome = `§7Biome: §r${getBiome}`
    const weather = `§7Weather: §r${getWeather} §7/ ${getTemperature}`
    const lightLevel = `§7Light Level: §r${getLight}`
    const moonPhase = `§7Moon Phase: §e${getMoonPhase()}`
    const spawnPoint = getSpawnPointAndDimension(getSpawnPoint)
    const widgets = []
    if (locationWidget) widgets.push(`${loc}`)
    if (dimensionWidget) widgets.push(`${dim}`)
    if (biomerWidget) widgets.push(`${biome}`)
    if (weatherWidget) widgets.push(`${weather}`)
    if (lightWidget) widgets.push(`${lightLevel}`)
    if (moontWidget) widgets.push(`${moonPhase}`)
    if (spawnpointWidget) widgets.push(`${spawnPoint}`)
    const newWidgets = widgets.join("\n")
    const myAlarms = Clock.getMyAlarms(userId)
    const ringingAlarms = myAlarms.filter(a => a.isRinging)
    const home = new ActionFormData()
        .title("SmartPhone")
        .body(`${owner}\n${timeStamp}\n${newWidgets}\n\n\n\n\n`)
        .button("Apps", "textures/coptaine_smartphone/icons/apps")
    for (const alarm of ringingAlarms) {
        const { name, time } = alarm
        home.button(`${name}\n${time}`, "textures/coptaine_smartphone/icons/alarm")
    }
    home.show(player).then((response) => {
        if (response.selection === 0) apps()
        for (const [index, alarm] of ringingAlarms.entries()) {
            const order = index
            if (response.selection - 1 === order) {
                player.runCommandAsync(`stopsound @s ${alarm.sound.value}`)
                alarm.isRinging = false
                Clock.serializeData(userId, myAlarms)
                smartphone(player)
            }
        }
    })

    function intro() {
        const form = new ActionFormData()
            .title("Smartphone")
            .body("§7Congratulations on your new smartphone! You've just unlocked a gateway to endless possibilities, where cutting-edge technology meets seamless functionality.\n\nDive into a vast ecosystem of apps, games and services. Welcome to a smarter, more connected world. Your new smartphone is here to empower, entertain, and elevate your everyday experiences. Enjoy the journey!\n\n")
            .button("Let's Go!")
        form.show(player).then((response) => {
            if (response.selection === 0) {
                PlayersRegistry.register(userId, userName)
                player.setDynamicProperty("stats:date_registered", getWorldDayTime())
                world.setDynamicProperty(`bank:${userId}`, 100)
                smartphone(player)
            }
        })
    }

    function apps(shortcut = undefined) {
        if (shortcut) {
            if (shortcut === "camera") return camera()
        }
        const form = new ActionFormData()
            .title("Applications")
            .button("Camera Pro", "textures/coptaine_smartphone/icons/camera")
            .button("Clock", "textures/coptaine_smartphone/icons/time")
            .button("Marketplace", "textures/coptaine_smartphone/icons/marketplace")
            .button("MineBank", "textures/coptaine_smartphone/icons/bank")
            .button("Minechat", "textures/coptaine_smartphone/icons/messaging")
            .button("Musicraft", "textures/coptaine_smartphone/icons/music")
            .button("Notes+", "textures/coptaine_smartphone/icons/notes")
            .button("Settings", "textures/coptaine_smartphone/icons/settings")
            .button("Statistics", "textures/coptaine_smartphone/icons/stats")
            .button("Weather", "textures/coptaine_smartphone/icons/weather")
            .button("§4 Back", "textures/coptaine_smartphone/icons/back")
        form.show(player).then((response) => {
            if (response.canceled) return
            if (response.selection === 0) camera()
            if (response.selection === 1) clock()
            if (response.selection === 2) marketplace()
            if (response.selection === 3) banking()
            if (response.selection === 4) messages()
            if (response.selection === 5) music()
            if (response.selection === 6) notes()
            if (response.selection === 7) settings()
            if (response.selection === 8) statistics()
            if (response.selection === 9) weather()
            if (response.selection === 10) smartphone(player)
        })

        function camera() {
            const cameras = Camera.getMyCameras(userId)
            const form = new ActionFormData()
                .title("Camera Pro")
                .body(`§aWelcome to Camera Pro! §7Monitor your property from anywhere and anytime, and instantly teleport to it.\n\n§oTIP: You can buy camera from Wandering Trader\n\n§eThe location where you will be teleported is where you were when you placed the camera\n\n§r§6Active Cameras: §r${cameras.length} / 5\n§7Teleport Cost: §g$30\n\n`)
                .button("Manage Cameras", "textures/coptaine_smartphone/icons/manage_cameras")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            for (const [index, camera] of cameras.entries()) {
                const name = camera.name
                const order = index + 1
                form.button(`${name}\n§8[Camera ${order}]`)
            }
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) manageCameras()
                if (response.selection === 1) apps()
                if (response.selection === 2) smartphone(player)
                for (const [index, camera] of cameras.entries()) {
                    const order = index
                    if (response.selection - 3 === order) {
                        const { tpLocation, location, rotation, dimension: cameraDimension } = camera
                        if (dimension.id !== cameraDimension.id) return dimensionError()
                        player.addTag("viewing_camera")
                        player.setDynamicProperty("camera_teleport", tpLocation)
                        player.sendMessage("§p[CAMERA PRO] §bUse again the smartphone to exit viewing.\nUse again the smartphone while sneaking to teleport.")
                        player.camera.setCamera("minecraft:free", { location: location, rotation: rotation })
                    }
                }
            })

            function dimensionError() {
                const form = new ActionFormData()
                    .title("Camera")
                    .body("§cAccess Error: This camera is in another dimension!\n\n")
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) camera()
                    if (response.selection === 1) smartphone(player)
                })
            }

            function manageCameras() {
                const cameras = Camera.getMyCameras(userId)
                const form = new ActionFormData()
                    .title("Manage Cameras")
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                for (const [index, camera] of cameras.entries()) {
                    const name = camera.name
                    const order = index + 1
                    form.button(`${name}\n§8[Camera ${order}]`)
                }
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) camera()
                    if (response.selection === 1) smartphone(player)
                    for (const [index, camera] of cameras.entries()) {
                        const order = index
                        if (response.selection - 2 === order) {
                            editCamera(camera, order)
                        }
                    }
                })

                function editCamera(camera, order) {
                    const { name, id, location, dimension: cameraDimension, isVisible } = camera
                    const visibilityAction = isVisible ? "Hide" : "Unhide"
                    const visibilityText = isVisible ? "§cVisible" : "§aInvisible"
                    const visibilityTexture = isVisible ? "hide" : "show"
                    const form = new ActionFormData()
                        .title(`${name}`)
                        .body(`§7Name: §r${name}\n§7Location: §r${location.x}, ${location.y}, ${location.z}\n§7Dimension: ${getDimensionName(cameraDimension)}\n§7Visibility: ${visibilityText}\n\n`)
                        .button("Rename", "textures/coptaine_smartphone/icons/edit")
                        .button(visibilityAction, `textures/coptaine_smartphone/icons/${visibilityTexture}`)
                        .button("Remove", "textures/coptaine_smartphone/icons/trash")
                        .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                        .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection < 3 && dimension.id !== cameraDimension.id) return dimensionError()
                        if (response.selection === 0) rename()
                        if (response.selection === 1) {
                            const cameras = dimension.getEntities({ type: "coptaine:camera" })
                            for (camera of cameras) {
                                if (camera.id === id) {
                                    if (isVisible) {
                                        camera.triggerEvent("e:hide")
                                        Camera.editCamera(userId, order, "isVisible", false)
                                    } else {
                                        camera.triggerEvent("e:show")
                                        Camera.editCamera(userId, order, "isVisible", true)
                                    }
                                    const updatedCamera = Camera.getMyCameras(userId)[order]
                                    editCamera(updatedCamera, order)
                                }
                            }
                        }
                        if (response.selection === 2) remove()
                        if (response.selection === 3) manageCameras()
                        if (response.selection === 4) smartphone(player)
                    })

                    function dimensionError() {
                        const form = new ActionFormData()
                            .title("Camera")
                            .body("§cAccess Error: This camera is in another dimension!\n\n")
                            .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                            .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            if (response.selection === 0) editCamera(camera, order)
                            if (response.selection === 1) smartphone(player)
                        })
                    }

                    function rename() {
                        const form = new ModalFormData()
                            .title(name)
                            .textField("Name:", "Leave blank to cancel")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            const [NAME] = response.formValues
                            if (!NAME) return editCamera(camera)
                            Camera.editCamera(userId, order, "name", NAME)
                            const updatedCamera = Camera.getMyCameras(userId)[order]
                            editCamera(updatedCamera, order)
                        })
                    }

                    function remove() {
                        const form = new ActionFormData()
                            .title(name)
                            .body("§7Do you want to remove this camera?\n\nNOTE: The camera will drop the item so you can use it again.\n\n")
                            .button("Remove")
                            .button("Cancel")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            if (response.selection === 0) {
                                Camera.removeCamera(userId, order)
                                const cameras = dimension.getEntities({ type: "coptaine:camera" })
                                for (camera of cameras) {
                                    if (camera.id === id) {
                                        camera.triggerEvent("e:despawn")
                                    }
                                }
                                manageCameras()
                            }
                            if (response.selection === 1) editCamera(camera, order)
                        })
                    }
                }
            }
        }

        function messages() {
            const form = new ActionFormData()
                .title("Minechat")
                .body("§aWelcome to Minechat! §7Stay connected with friends and enjoy seamless conversations on the go.\n\n")
                .button("Compose", "textures/coptaine_smartphone/icons/compose")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")

            const myConversations = Conversations.getMyConversations(userId)
            for (const conversation of myConversations) {
                const receiver = Conversations.getMessage(userId, extractId(conversation, userId)).receiverName
                const message = Conversations.getMessage(userId, extractId(conversation, userId)).newMessage
                form.button(`${receiver}\n${message}`)
            }

            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) compose()
                if (response.selection === 1) apps()
                if (response.selection === 2) smartphone(player)
                for (const [index, conversation] of myConversations.entries()) {
                    const order = index
                    if (response.selection - 3 === order) message(`${extractId(conversation, userId)}`)
                }
            })

            function compose() {
                const { ids, names } = PlayersRegistry.getRegistered()
                const form = new ModalFormData()
                    .dropdown("Players:", names)
                    .textField("Message:", "Leave blank to cancel")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    const [RECEIVER, MESSAGE] = response.formValues
                    const targetId = ids[RECEIVER]
                    const targetName = names[RECEIVER]

                    if (userId === targetId || !MESSAGE) return messages()
                    if (Conversations.blockStatus(targetId, userId)) return youAreBlocked("compose", targetId)
                    Conversations.sendMessage(userName, targetName, userId, targetId, MESSAGE)
                    messages()
                })
            }

            function message(receiverId) {
                const data = Conversations.getMessage(userId, receiverId)
                const body = data.message
                const receiver = data.receiverName
                const muteOrUnmute = Conversations.muteStatus(userId, receiverId) ? "Unmute" : "Mute"
                const blockOrUnblock = Conversations.blockStatus(userId, receiverId) ? "Unblock" : "Block"
                const form = new ActionFormData()
                    .title(receiver)
                    .body(body)
                    .button("Reply", "textures/coptaine_smartphone/icons/reply")
                    .button(muteOrUnmute, "textures/coptaine_smartphone/icons/mute")
                    .button(blockOrUnblock, "textures/coptaine_smartphone/icons/block")
                    .button("Delete", "textures/coptaine_smartphone/icons/trash")
                    .button("Close", "textures/coptaine_smartphone/icons/close")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) reply()
                    if (response.selection === 1) mute(muteOrUnmute)
                    if (response.selection === 2) block(blockOrUnblock)
                    if (response.selection === 3) deleteConfirmation()
                    if (response.selection === 4) messages()
                })

                function reply() {
                    if (Conversations.blockStatus(receiverId, userId)) return youAreBlocked("reply", receiverId)
                    const unblockWarning = Conversations.blockStatus(userId, receiverId) ? "\n\n§cReplying to this conversation will unblock the user!" : ""
                    const form = new ModalFormData()
                        .title(`Replying to ${receiver}`)
                        .textField(`Message:${unblockWarning}`, "Leave blank to cancel")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        const [MESSAGE] = response.formValues

                        if (!MESSAGE) return message(receiverId)
                        Conversations.sendMessage(userName, receiver, userId, receiverId, MESSAGE)
                        if (Conversations.blockStatus(userId, receiverId)) Conversations.unblockConversation(userId, receiverId)
                        message(receiverId)
                    })
                }

                function deleteConfirmation() {
                    const form = new ActionFormData()
                        .title(`${receiver}`)
                        .body("§7Delete this conversation? This is permanent and can't be undone.\n\n")
                        .button("Delete")
                        .button("Cancel")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) {
                            world.setDynamicProperty(`conversation:${userId}_${receiverId}`, undefined)
                            messages()
                        }
                        if (response.selection === 1) message(receiverId)
                    })
                }

                function mute(mute) {
                    const confrimationMessage = [
                        "§7You will no longer receive notifications from this user. Do you want to continue?\n\n",
                        "§7You will receive notifications from this user. Do you want to continue?\n\n"
                    ]
                    const form = new ActionFormData()
                        .title(`${receiver}`)
                        .body(`${mute === "Mute" ? confrimationMessage[0] : confrimationMessage[1]}`)
                        .button("Yes")
                        .button("Cancel")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) {
                            if (mute === "Mute") {
                                Conversations.muteConversation(userId, receiverId)
                            } else {
                                Conversations.unmuteConversation(userId, receiverId)
                            }
                            message(receiverId)
                        }
                        if (response.selection === 1) message(receiverId)
                    })
                }

                function block(block) {
                    const confrimationMessage = [
                        "§7You will no longer receive messages from this user. Do you want to continue?\n\n",
                        "§7You will receive messages from this user. Do you want to continue?\n\n"
                    ]
                    const form = new ActionFormData()
                        .title(`${receiver}`)
                        .body(`${block === "Block" ? confrimationMessage[0] : confrimationMessage[1]}`)
                        .button("Yes")
                        .button("Cancel")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) {
                            if (block === "Block") {
                                Conversations.blockConversation(userId, receiverId)
                            } else {
                                Conversations.unblockConversation(userId, receiverId)
                            }
                            message(receiverId)
                        }
                        if (response.selection === 1) message(receiverId)
                    })
                }
            }

            function youAreBlocked(method, receiverId) {
                const composeOrReply = [
                    `§cYou cannot send message to this user!\n\n`,
                    `§cYou cannot reply to this conversation!\n\n`
                ]
                const form = new ActionFormData()
                    .title("Minechat")
                    .body(`${method === "compose" ? composeOrReply[0] : composeOrReply[1]}`)
                    .button("Close")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) {
                        if (method === "compose") return messages()
                        message(receiverId)
                    }
                })
            }
        }

        function notes() {
            const myNotes = Notes.getMyNotes(userId)
            const form = new ActionFormData()
                .title("Notes+")
                .body("§aWelcome to Notes+! §7Capture your thoughts, organize your ideas, access them anytime, and print them conveniently.\n\n")
                .button("Compose", "textures/coptaine_smartphone/icons/notes_compose")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            for (const note of myNotes) {
                const { title, message } = note
                form.button(`${title}\n${message}`)
            }
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) addNote()
                if (response.selection === 1) apps()
                if (response.selection === 2) smartphone(player)
                for (const [index, note] of myNotes.entries()) {
                    const order = index
                    if (response.selection - 3 === order) openNote(note, order)
                }
            })

            function addNote() {
                const form = new ModalFormData()
                    .title("Notes+")
                    .textField("Title:", "Leave blank to cancel")
                    .textField("§o§7TIP: Type | to create a new line\n\n§rMessage:", "Text")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    const [TITLE, MESSAGE] = response.formValues
                    if (!TITLE) return notes()
                    Notes.addNote(userId, TITLE, splitString(MESSAGE))
                    notes()
                })
            }

            function openNote(note, order) {
                const { title, message } = note
                const form = new ActionFormData()
                    .title(`${title}`)
                    .body(`${message}`)
                    .button("Edit", "textures/coptaine_smartphone/icons/edit")
                    .button("Print", "textures/coptaine_smartphone/icons/print")
                    .button("Delete", "textures/coptaine_smartphone/icons/trash")
                    .button("Close", "textures/coptaine_smartphone/icons/close")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) editNote()
                    if (response.selection === 1) printNote()
                    if (response.selection === 2) deleteConfirmation()
                    if (response.selection === 3) notes()
                })

                function editNote() {
                    const formatedMessage = message.replace("\n", "|")
                    const form = new ModalFormData()
                        .title(title)
                        .textField("Title:", title, title)
                        .textField("§o§7TIP: Type | to create a new line\n\n§rMessage:", formatedMessage, formatedMessage)
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        const [TITLE, MESSAGE] = response.formValues
                        Notes.editNote(userId, order, TITLE, splitString(MESSAGE))
                        notes()
                    })
                }

                function printNote() {
                    const form = new ModalFormData()
                        .title("Print+")
                        .slider("§7Printed notes can be placed in item frames and holograms.\n\n§oTIP: Select 0 to cancel\n\n§rCopies", 0, 64, 1)
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        const [AMOUNT] = response.formValues
                        if (AMOUNT === 0) return openNote(note)
                        const item = new ItemStack("coptaine:printed_note", AMOUNT)
                        item.nameTag = `${title}\n${message}`
                        player.getComponent("inventory").container.addItem(item)
                        openNote(note, order)
                    })
                }

                function deleteConfirmation() {
                    const form = new ActionFormData()
                        .title(`${title}`)
                        .body("§7Delete this note? This is permanent and can't be undone.\n\n")
                        .button("Delete")
                        .button("Cancel")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) {
                            Notes.removeNote(userId, note)
                            notes()
                        }
                        if (response.selection === 1) openNote(note)
                    })
                }
            }
        }

        function banking() {
            const balance = Bank.balance(userId)
            const form = new ActionFormData()
                .title("MineBank")
                .body(`§aWelcome to your MineBank! §7As an owner of the Ultimate Smartphone, you are also part of the §gOnline Currency System§7. Manage your accounts, transfer funds, and monitor your transactions securely and conveniently.\n\nAccount Owner: §r${userName}\n§7Available Balance: §g$${balance}\n\n`)
                .button("Transfer", "textures/coptaine_smartphone/icons/transfer")
                .button("Vault", "textures/coptaine_smartphone/icons/vault")
                .button("Transactions", "textures/coptaine_smartphone/icons/transactions")
                .button("§q Refresh", "textures/coptaine_smartphone/icons/refresh")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) transfer()
                if (response.selection === 1) vault()
                if (response.selection === 2) transactions()
                if (response.selection === 3) banking()
                if (response.selection === 4) apps()
                if (response.selection === 5) smartphone(player)
            })

            function transfer() {
                const { ids, names } = PlayersRegistry.getRegistered()
                const form = new ModalFormData()
                    .title("Transfer")
                    .dropdown("Players:", names)
                    .textField("Amount:", "Leave blank to cancel")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    const [RECEIVER, AMOUNT] = response.formValues
                    const targetId = ids[RECEIVER]
                    const receiverName = names[RECEIVER]
                    const checkedAmount = isNonEmptyString(AMOUNT) ? AMOUNT : 0

                    if (checkedAmount == 0) return banking() // Should always be "=="
                    if (!isNumeric(checkedAmount) || targetId === userId || checkedAmount < 0 || balance < checkedAmount) return transferFailed()
                    Bank.transfer(Number(checkedAmount), userId, targetId, userName, receiverName)
                    transferSuccessful(checkedAmount, receiverName)
                })

                function transferSuccessful(amount, receiver) {
                    const form = new ActionFormData()
                        .title("Bank")
                        .body(`Successfully transfered §g$${amount} §rto ${receiver}!\n\n`)
                        .button("Close")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) banking()
                    })
                }

                function transferFailed() {
                    const form = new ActionFormData()
                        .title("Bank")
                        .body(`§cTransfer failed: Invalid amount or recipient!\n\n`)
                        .button("Close")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) banking()
                    })
                }
            }

            function vault() {
                const form = new ActionFormData()
                    .title("Vault")
                    .body(`§7Welcome §r${userName} §7to your vault! Here you can store your most precious items safely!\n\n`)
                    .button("Open", "textures/coptaine_smartphone/icons/vault_open")
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.selection === 0) open()
                    if (response.selection === 1) banking()
                    if (response.selection === 2) smartphone(player)
                })

                function open() {
                    if (dimension.id !== "minecraft:overworld") return player.sendMessage("§g[Bank] §cCannot access the vault in this dimension!")
                    Bank.openVault(player)
                    system.runTimeout(() => {
                        player.runCommandAsync(`ride @s start_riding @e[type=coptaine:vault,c=1,tag=tag:${userId}]`)
                        player.sendMessage("§g[BANK] §bOpen your inventory to access the vault")
                    }, 10)
                }
            }

            function transactions() {
                const message = Bank.transactions(userId) ?? ""
                const form = new ActionFormData()
                    .title("Transactions")
                    .body(`${message}`)
                    .button("Close")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) banking()
                })
            }
        }

        function marketplace() {
            const form = new ActionFormData()
                .title("Marketplace")
                .body(`§aWelcome to the Marketplace! §7Explore a wide range of items, buy or trade with other players, and enjoy seamless shopping at your fingertips.\n\n§oTIP: You can sell items to Wandering Trader!\n\n§r§7Money: §g$${Bank.balance(userId)}\n\n`)
                .button("Construction", "textures/coptaine_smartphone/icons/construction")
                .button("Equipment", "textures/coptaine_smartphone/icons/equipment")
                .button("Items", "textures/coptaine_smartphone/icons/items")
                .button("Nature", "textures/coptaine_smartphone/icons/nature")
                .button("Wandering Trader", "textures/coptaine_smartphone/icons/trader")
                .button("Trade with Players", "textures/coptaine_smartphone/icons/players")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) buy(Marketplace.construction(), "Construction")
                if (response.selection === 1) buy(Marketplace.equipment(), "Equipment")
                if (response.selection === 2) buy(Marketplace.items(), "Items")
                if (response.selection === 3) buy(Marketplace.nature(), "Nature")
                if (response.selection === 4) wanderingTrader()
                if (response.selection === 5) {
                    if (dimension.id !== "minecraft:overworld") return player.sendMessage("§a[MARKETPLACE] §cCannot trade with other players in this dimension!")
                    playersTrades()
                }
                if (response.selection === 6) apps()
                if (response.selection === 7) smartphone(player)
            })

            function buy(items, marketName) {
                const balance = Bank.balance(userId)
                const form = new ActionFormData()
                    .title(marketName)
                    .body(`§7Money: §g$${balance}\n\n`)
                for (const item of items) {
                    const { id, name, amount, price } = item
                    const icon = id.split(":")[1]
                    form.button(`${name} x${amount}\n§oPrice: §7§c$${price}`, `textures/coptaine_smartphone/icons/${icon}`)
                }
                form.button("§4 Back", "textures/coptaine_smartphone/icons/back")
                form.button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    for (const [index, item] of items.entries()) {
                        const { id, name, amount, price } = item
                        const order = index
                        if (response.selection === order) {
                            if (balance < price) return purchase("failed", name, `x${amount}`, marketName, price, balance)
                            const getItem = new ItemStack(id, amount)
                            getInventory(player).container.addItem(getItem)
                            Bank.transfer(price, userId, 0, userName, "§nTrader")
                            purchase("succeeded", name, `x${amount}`, marketName, price, balance)
                        }
                    }
                    if (response.selection === items.length) marketplace()
                    if (response.selection === items.length + 1) smartphone(player)
                })
            }

            function wanderingTrader(shortcut = undefined) {
                if (shortcut) {
                    if (shortcut === "buy") return buy()
                    if (shortcut === "sell") return sell()
                }
                const balance = Bank.balance(userId)
                const form = new ActionFormData()
                    .title("Wandering Trader")
                    .body(`§7I sell unique and special items, and I also buy various items!\n\nMoney: §g$${balance}\n\n`)
                    .button("Buy", "textures/coptaine_smartphone/icons/buy")
                    .button("Sell", "textures/coptaine_smartphone/icons/sell")
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) buy(Marketplace.wanderingTrades(), "modifyCancel")
                    if (response.selection === 1) sell()
                    if (response.selection === 2) marketplace()
                    if (response.selection === 3) smartphone(player)
                })

                function buy() {
                    const balance = Bank.balance(userId)
                    const items = Marketplace.wanderingTrades()
                    const form = new ActionFormData()
                        .title("Wandering Trader")
                        .body(`Money: §g$${balance}\n\n`)
                        .button("Camera", "textures/coptaine_smartphone/items/camera")
                        .button("Hologram Projector", "textures/coptaine_smartphone/items/holo_billboard")
                        .button("Enchanted Bow", "textures/coptaine_smartphone/icons/bow")
                        .button("Enchanted Chainmail Boots", "textures/coptaine_smartphone/icons/chainmail_boots")
                        .button("Enchanted Crossbow", "textures/coptaine_smartphone/icons/crossbow")
                        .button("Enchanted Diamond Leggings", "textures/coptaine_smartphone/icons/diamond_leggings")
                        .button("Enchanted Diamond Sword", "textures/coptaine_smartphone/icons/diamond_sword")
                        .button("Enchanted Elytra", "textures/coptaine_smartphone/icons/elytra")
                        .button("Enchanted Fishing Rod", "textures/coptaine_smartphone/icons/fishing_rod")
                        .button("Enchanted Golden Axe", "textures/coptaine_smartphone/icons/gold_axe")
                        .button("Enchanted Iron Helmet", "textures/coptaine_smartphone/icons/iron_helmet")
                        .button("Enchanted Iron Pickaxe", "textures/coptaine_smartphone/icons/iron_pickaxe")
                        .button("Enchanted Mace", "textures/coptaine_smartphone/icons/mace")
                        .button("Enchanted Netherite Chestplate", "textures/coptaine_smartphone/icons/netherite_chestplate")
                        .button("Enchanted Shield", "textures/coptaine_smartphone/icons/shield")
                        .button("Enchanted Trident 1", "textures/coptaine_smartphone/icons/trident")
                        .button("Enchanted Trident 2", "textures/coptaine_smartphone/icons/trident")
                        .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                        .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        for (const [index, item] of items.entries()) {
                            const order = index
                            if (response.selection === order) checkItem(item)
                        }
                        if (response.selection === 17) wanderingTrader()
                        if (response.selection === 18) smartphone(player)
                    })

                    function checkItem(item) {
                        const balance = Bank.balance(userId)
                        const { name, description, price, structure } = item
                        const form = new ActionFormData()
                            .title(`Money: $${balance}`)
                            .body(`§7Name: §b${name}\n${description}§7Price: §g$${price}`)
                            .button("Buy", "textures/coptaine_smartphone/icons/buy")
                            .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                            .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            if (response.selection === 0) {
                                if (balance < price) return purchase("failed", name, 1, "Wandering Trader", price, balance)
                                Bank.transfer(price, userId, 0, userName, "§nWandering Trader")
                                purchase("succeeded", name, "x1", "Wandering Trader", price, balance)
                                if (structure === "camera") {
                                    getInventory(player).container.addItem(new ItemStack("coptaine:camera"))
                                } else if (structure === "holo_billboard") {
                                    getInventory(player).container.addItem(new ItemStack("coptaine:holo_billboard"))
                                } else {
                                    world.structureManager.place(structure, dimension, location)
                                }
                            }
                            if (response.selection === 1) buy()
                            if (response.selection === 2) smartphone(player)
                        })
                    }
                }

                function sell() {
                    const balance = Bank.balance(userId)
                    const playerSellable = Marketplace.player()
                    const { container, items, itemSlots } = getInventory(player)

                    const filteredItems = items.map(i => {
                        const sellable = playerSellable.find(sell => sell.id === i.typeId)
                        if (sellable) {
                            return {
                                id: sellable.id,
                                price: sellable.price,
                                name: sellable.name,
                                amount: i.amount,
                                slot: itemSlots[items.findIndex(s => s == i)]
                            }
                        }
                    }).filter(i => i !== undefined)
                    const trades = filteredItems.map((i) => {
                        return `${i.name} x${i.amount}: $${i.price} each`;
                    })
                    trades.push("Cancel")
                    const form = new ModalFormData()
                        .title("Wandering Trader")
                        .dropdown(`§7Money: §g$${balance}\n§7§oTIP: You can cancel by selecting Cancel\n\n§rItems:`, trades)
                        .slider("Amount", 1, 64, 1)
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        const [ITEM, QUANTITY] = response.formValues
                        if (trades[ITEM] === "Cancel") return wanderingTrader()

                        const selectedItem = filteredItems[ITEM]
                        const { slot, amount, price, name } = selectedItem
                        const finalAmount = QUANTITY >= amount ? amount : QUANTITY
                        const totalPrice = price * finalAmount
                        const item = QUANTITY >= amount ? "minecraft:air" : selectedItem.id
                        const getItem = new ItemStack(item, Math.max(amount - QUANTITY, 1))
                        Bank.transfer(Number(totalPrice), 0, userId, "§nWandering Trader", userName)
                        purchase("sold", name, finalAmount, "Wandering Trader", totalPrice, balance)
                        container.setItem(slot, item === "minecraft:air" ? null : getItem)
                    })
                }
            }

            function playersTrades() {
                const balance = Bank.balance(userId)
                const shops = Marketplace.getShops()
                const myShop = `shop:${userId}`
                const form = new ActionFormData()
                    .title("Player Trades")
                    .body(`§7Buy or sell with other players!\n\n§r§7Money: §g$${balance}\n\n`)
                    .button("My Shop", "textures/coptaine_smartphone/icons/marketplace")
                    .button("§q Refresh", "textures/coptaine_smartphone/icons/refresh")
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                for (const shop of shops) {
                    const message = Marketplace.getShopData(shop).message
                    const owner = Marketplace.getShopData(shop).owner
                    form.button(`${message}\n§8§oOwner: ${owner}`)
                }

                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) {
                        const getMyShop = Marketplace.getMyShop(userId)
                        if (!getMyShop.length) return openAShop()
                        manageMyShop()
                    }
                    if (response.selection === 1) playersTrades()
                    if (response.selection === 2) marketplace()
                    if (response.selection === 3) smartphone(player)
                    for (const [index, shop] of shops.entries()) {
                        const order = index
                        const data = Marketplace.getShopData(shop)
                        if (response.selection - 4 === order) {
                            if (userId === data.id) return manageMyShop()
                            visitShop(data)
                            Marketplace.updateShopStats(`shop_stats:${data.id}`, "visits")
                        }
                    }
                })

                function openAShop() {
                    const { container, items, itemSlots } = getInventory(player)
                    const amounts = items.map(i => i.amount)
                    const trades = items.map(i => `${i.typeId} x${i.amount}`)
                    trades.push("Cancel")
                    const form = new ModalFormData()
                        .title("My Shop")
                        .textField("§7§oTIP: You can cancel by selecting Cancel\n\n§rMessage:", "Your shop description")
                        .dropdown("Item:", trades)
                        .textField("Selling Price:", "Price")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        const [MESSAGE, ITEM, PRICE] = response.formValues
                        if (trades[ITEM] === "Cancel") return playersTrades()
                        const price = !isNonEmptyString(PRICE) || !isNumeric(PRICE) ? 0 : PRICE
                        const amount = amounts[ITEM]
                        const itemSlot = itemSlots[ITEM]
                        const item = container.getItem(itemSlot)
                        const itemEnchantments = item.getComponent("enchantable")?.getEnchantments()
                        const itemNameTag = item.nameTag
                        const itemLores = item.getLore()
                        const itemData = [item.typeId, itemNameTag, itemEnchantments, itemLores, price, amount]
                        const shopData = [myShop, MESSAGE, userName]
                        Marketplace.addItem(player, item, itemSlot)
                        Marketplace.editMyShop(shopData, itemData, userId)
                        manageMyShop()
                    })
                }

                function manageMyShop() {
                    const balance = Bank.balance(userId)
                    const { items, message } = Marketplace.getShopData(myShop)
                    const stats = Marketplace.getShopStats(`shop_stats:${userId}`)
                    const { visits, sold } = stats
                    const totalSales = stats.totalSales
                    const color = items.length < 5 ? "§2" : "§c§l"
                    const form = new ActionFormData()
                        .title("My Shop")
                        .body(`${message}\n\n§7Visits: §r${visits}\n§7Sold: §r${sold}\n§7Total Sales: §a$${totalSales}\n§7Money: §g$${balance}\n\n`)
                        .button("Edit Message", "textures/coptaine_smartphone/icons/edit")
                        .button(`Add Item\n§o${color}<${items.length} / 5>`, "textures/coptaine_smartphone/icons/add_item")
                        .button("Delete", "textures/coptaine_smartphone/icons/trash")
                        .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                        .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                    for (const item of items) {
                        const { id, amount, price } = item
                        form.button(`${id} x${amount}\n§8§oPrice: §c$${price}`)
                    }
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) editMessage()
                        if (response.selection === 1) {
                            if (items.length < 5) return addItem()
                            return itemLimit()
                        }
                        if (response.selection === 2) deleteConfirmation()
                        if (response.selection === 3) playersTrades()
                        if (response.selection === 4) smartphone(player)
                        for (const [index, item] of items.entries()) {
                            const order = index
                            const slot = item.slot
                            if (response.selection - 5 === order) editItem(item, slot, order)
                        }
                    })

                    function itemLimit() {
                        const form = new ActionFormData()
                            .title("My Shop")
                            .body("§cMax items is reached (5/5)!\n\n")
                            .button("Back")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            if (response.selection === 0) manageMyShop()
                        })
                    }

                    function deleteConfirmation() {
                        const form = new ActionFormData()
                            .title("My Shop")
                            .body("§7Do you want to delete your shop?\n\n")
                            .button("Delete")
                            .button("Cancel")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            if (response.selection === 0) {
                                for (const item of items) {
                                    Marketplace.retrieveItem(getInventory(player).container, item.slot, 0)
                                }
                                world.setDynamicProperty(myShop, undefined)
                                world.setDynamicProperty(`shop_stats${userId}`, undefined)
                                playersTrades()

                            }
                            if (response.selection === 1) manageMyShop()
                        })
                    }

                    function editMessage() {
                        const form = new ModalFormData()
                            .title("My Shop")
                            .textField("§7§oTIP: Leave blank to cancel\n\n§rMessage:", "Your shop description")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            const [MESSAGE] = response.formValues
                            if (!MESSAGE) return manageMyShop()
                            const shopData = [myShop, MESSAGE, userName]
                            Marketplace.editMyShop(shopData, undefined, userId)
                            manageMyShop()
                        })
                    }

                    function addItem() {
                        const message = Marketplace.getShopData(myShop).message
                        const { container, items, itemSlots } = getInventory(player)
                        const amounts = items.map(i => i.amount)
                        const trades = items.map(i => `${i.typeId} x${i.amount}`)
                        trades.push("Cancel")
                        const form = new ModalFormData()
                            .title("My Shop")
                            .dropdown("§7§oTIP: You can cancel by selecting Cancel\n\n§rItem:", trades)
                            .textField("Selling Price:", "Price")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            const [ITEM, PRICE] = response.formValues
                            if (trades[ITEM] === "Cancel") return manageMyShop()
                            const price = !isNonEmptyString(PRICE) || !isNumeric(PRICE) ? 0 : PRICE
                            const amount = amounts[ITEM]
                            const itemSlot = itemSlots[ITEM]
                            const item = container.getItem(itemSlot)
                            const itemEnchantments = item.getComponent("enchantable")?.getEnchantments()
                            const itemNameTag = item.nameTag
                            const itemLore = item.getLore()
                            const itemData = [item.typeId, itemNameTag, itemEnchantments, itemLore, price, amount]
                            const shopData = [myShop, message, userName]
                            Marketplace.addItem(player, item, itemSlot)
                            Marketplace.editMyShop(shopData, itemData, userId)
                            manageMyShop()
                        })
                    }

                    function editItem(data, slot, shopSlot) {
                        const message = Marketplace.getItemInfo(data)
                        const form = new ActionFormData()
                            .title("Edit Item")
                            .body(message)
                            .button("Remove", "textures/coptaine_smartphone/icons/sell")
                            .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                            .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            if (response.selection === 0) {
                                const message = Marketplace.getShopData(myShop).message
                                const shopData = [myShop, message, userName]
                                Marketplace.retrieveItem(getInventory(player).container, slot, 0)
                                Marketplace.editMyShop(shopData, shopSlot, userId)
                                manageMyShop()
                            }
                            if (response.selection === 1) manageMyShop()
                            if (response.selection === 2) smartphone(player)
                        })
                    }
                }

                function visitShop(data) {
                    const balance = Bank.balance(userId)
                    const { name: shopName, owner: shoptitle, message: shopMessage, items: shopItems, id: shopOwnerId } = data
                    const form = new ActionFormData()
                        .title(`${shoptitle}'s Shop`)
                        .body(`${shopMessage}\n\n§7Money: §g$${balance}\n\n`)
                        .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                        .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                    for (const item of shopItems) {
                        const { id, amount, price } = item
                        form.button(`${id} x${amount}\n§8§oPrice: §c$${price}`)
                    }
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) playersTrades()
                        if (response.selection === 1) smartphone(player)
                        for (const [index, item] of shopItems.entries()) {
                            const order = index
                            const slot = item.slot
                            if (response.selection - 2 === order) checkItem(shopItems[order], slot, shoptitle, order)
                        }
                    })

                    function checkItem(itemData, itemSlot, ownerName, shopSlot) {
                        const { id, amount, price } = itemData
                        const message = Marketplace.getItemInfo(itemData)
                        const form = new ActionFormData()
                            .title(`Money: $${balance}`)
                            .body(message)
                            .button("Buy", "textures/coptaine_smartphone/icons/buy")
                            .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                            .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            if (response.selection === 0) {
                                if (balance < Number(price)) return purchase("failed")
                                Bank.transfer(Number(price), userId, shopOwnerId, userName, ownerName)
                                purchase("succeeded")
                                const getData = Marketplace.getShopData(shopName)
                                const shopData = [shopName, getData.message, getData.owner]
                                Marketplace.retrieveItem(getInventory(player).container, itemSlot, 1, shopOwnerId)
                                Marketplace.editMyShop(shopData, shopSlot, getData.id)
                                Marketplace.updateShopStats(`shop_stats:${shopOwnerId}`, "sold")
                                Marketplace.updateShopStats(`shop_stats:${shopOwnerId}`, "totalSales", price)
                            }
                            if (response.selection === 1) {
                                const refreshedData = Marketplace.getShopData(shopName)
                                visitShop(refreshedData)
                            }
                            if (response.selection === 2) smartphone(player)
                        })

                        function purchase(result) {
                            const failedMessage = `§cTransaction failed: Not enough money!\n§7Money: §g$${balance}\n§7You need §g$${price}\n\n`
                            const succeededMessage = `${id} x${amount}`
                            const chooseMessage = result === "succeeded" ? succeededMessage : failedMessage
                            const form = new ActionFormData()
                                .title(`${owner}'s Shop`)
                                .body(chooseMessage)
                                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                            form.show(player).then((response) => {
                                if (response.canceled) return
                                if (response.selection === 0) {
                                    const refreshedData = Marketplace.getShopData(shopName)
                                    visitShop(refreshedData)
                                }
                                if (response.selection === 1) smartphone(player)
                            })
                        }
                    }
                }
            }

            function purchase(result, name, amount, shopName, price, balance) {
                const message = [
                    `§aSuccessfully sold ${name} ${amount} for $${price}!\n\n`,
                    `§aSuccessfully bought ${name} ${amount}!\n\n`,
                    `§cTransaction failed: Not enough money!\n§7Money: §g$${balance}\n§7You need §g$${price}\n\n`
                ]
                const chooseMessage = result === "sold" ? message[0] : (result === "succeeded" ? message[1] : message[2])
                const form = new ActionFormData()
                    .title(shopName)
                    .body(chooseMessage)
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("Marketplace", "textures/coptaine_smartphone/icons/marketplace")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) {
                        if (shopName === "Wandering Trader" && result !== "sold") wanderingTrader("buy")
                        if (shopName === "Wandering Trader" && result === "sold") wanderingTrader("sell")
                        if (shopName === "Construction") buy(Marketplace.construction(), "Construction")
                        if (shopName === "Items") buy(Marketplace.items(), "Items")
                        if (shopName === "Equipment") buy(Marketplace.equipment(), "Equipment")
                        if (shopName === " Nature") buy(Marketplace.nature(), "Nature")
                    }
                    if (response.selection === 1) marketplace()
                    if (response.selection === 2) smartphone(player)
                })
            }
        }

        function music() {
            const playing = player.getDynamicProperty("music") ?? ""
            const form = new ActionFormData()
                .title("Musicraft")
                .body(`§aWelcome to Musicraft! §7Enjoy your favorite tracks anytime, anywhere!\n\nNow playing: §b${playing}\n\n`)
                .button("Select Music", "textures/coptaine_smartphone/icons/music")
                .button("Stop", "textures/coptaine_smartphone/icons/stop")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) selectMusic()
                if (response.selection === 1) {
                    player.setDynamicProperty("music", undefined)
                    player.stopMusic()
                    music()
                }
                if (response.selection === 2) apps()
                if (response.selection === 3) smartphone(player)
            })

            function selectMusic() {
                const filteredMusicTracks = musicTracks.map(n => n[0].toUpperCase() + n.slice(1))
                const form = new ModalFormData()
                    .title("Music")
                    .dropdown("Tracks:", filteredMusicTracks)
                    .toggle("Loop")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    const [MUSIC, LOOP] = response.formValues
                    player.playMusic(`coptaine_us.music.${musicTracks[MUSIC]}`, { loop: LOOP })
                    player.setDynamicProperty("music", filteredMusicTracks[MUSIC])
                    music()
                })
            }
        }

        function weather() {
            const balance = Bank.balance(userId)
            const { getWeather, getTemperature } = getPlayerStats(player)
            const weatherString = `§7Weather: §r${getWeather} §7/ ${getTemperature}\n§7Moon Phase: §e${getMoonPhase()}\n\n`
            const form = new ActionFormData()
                .title("Weather")
                .body(`§aWelcome to Weather! §7Get accurate forecasts and plan your day with confidence.\n\n${weatherString}§7As an owner of the Ultimate Smartphone, you can also enjoy exclusive services from the weather bureau, allowing you to change the weather for a fee (§g$100§7).\n\n`)
                .button("Clear", "textures/coptaine_smartphone/icons/weather_clear")
                .button("Rain", "textures/coptaine_smartphone/icons/weather_rain")
                .button("Thunderstorm", "textures/coptaine_smartphone/icons/weather_thunderstorm")
                .button("§q Refresh", "textures/coptaine_smartphone/icons/refresh")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) changeWeather("Clear")
                if (response.selection === 1) changeWeather("Rain")
                if (response.selection === 2) changeWeather("Thunder")
                if (response.selection === 3) {
                    weatherAtPosition(player)
                    system.runTimeout(() => {
                        weather()
                    }, 5)
                }
                if (response.selection === 4) apps()
                if (response.selection === 5) smartphone(player)
            })

            function changeWeather(getWeather) {
                if (balance < 100) return weatherChanged("failed", getWeather)
                dimension.setWeather(getWeather)
                weatherChanged("succeeded", getWeather)
                Bank.transfer(100, userId, 0, userName, "§nWeather Bureau")
            }

            function weatherChanged(result, getWeather) {
                const message = [
                    `§aSuccessfully changed weather to ${getWeather}!\n\n`,
                    `§cFailed to change the weather: Not enough money!\n§7Money: §g$${balance}\n§7You need §g$100\n\n`
                ]
                const form = new ActionFormData()
                    .title("Weather")
                    .body(`${result === "succeeded" ? message[0] : message[1]}`)
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.selection === 0) weather()
                    if (response.selection === 1) smartphone(player)
                })
            }
        }

        function clock() {
            const balance = Bank.balance(userId)
            const timeStamp = getWorldDayTime()
            const form = new ActionFormData()
                .title("Clock")
                .body(`§aWelcome to Clock! §7Track your progress and manage your schedule effortlessly.\n\n${timeStamp}\n\n§7Changing time is strictly forbidden but with the right price (§g$10000§7), anything is possible.\n\n`)
                .button("Alarms", "textures/coptaine_smartphone/icons/alarm")
                .button("Sunrise", "textures/coptaine_smartphone/icons/sunrise")
                .button("Day", "textures/coptaine_smartphone/icons/day")
                .button("Noon", "textures/coptaine_smartphone/icons/noon")
                .button("Sunset", "textures/coptaine_smartphone/icons/sunset")
                .button("Night", "textures/coptaine_smartphone/icons/night")
                .button("Midnight", "textures/coptaine_smartphone/icons/midnight")
                .button("§q Refresh", "textures/coptaine_smartphone/icons/refresh")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) alarms()
                if (response.selection === 1) changeTime("sunrise")
                if (response.selection === 2) changeTime("day")
                if (response.selection === 3) changeTime("noon")
                if (response.selection === 4) changeTime("sunset")
                if (response.selection === 5) changeTime("night")
                if (response.selection === 6) changeTime("midnight")
                if (response.selection === 7) time()
                if (response.selection === 8) apps()
                if (response.selection === 9) smartphone(player)
            })

            function alarms() {
                const alarmSoundTexts = alarmSounds.map(s => s.name)
                const alarmSoundValues = alarmSounds.map(s => s.value)
                const myAlarms = Clock.getMyAlarms(userId)
                const form = new ActionFormData()
                    .title("Alarms")
                    .button("Add", "textures/coptaine_smartphone/icons/add_alarm")
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                for (const alarm of myAlarms) {
                    const { name, time, isActive } = alarm
                    const statusTexture = isActive ? "active" : "inactive"
                    form.button(`${name}\n${time}`, `textures/coptaine_smartphone/icons/${statusTexture}`)
                }
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) addAlarm()
                    if (response.selection === 1) clock()
                    if (response.selection === 2) smartphone(player)
                    for (const [index, alarm] of myAlarms.entries()) {
                        const order = index
                        if (response.selection - 3 === order) {
                            player.runCommandAsync(`stopsound @s ${alarm.sound.value}`)
                            alarm.isRinging = false
                            Clock.serializeData(userId, myAlarms)
                            manageAlarm(alarm, order)
                        }
                    }
                })

                function addAlarm() {
                    const form = new ModalFormData()
                        .title("Add Alarm")
                        .textField("Name:", "Leave blank to cancel")
                        .slider("Hour", 0, 23, 1)
                        .slider("Minute", 0, 59, 1)
                        .dropdown("Sound:", alarmSoundTexts)
                        .toggle("Repeat", false)
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        const [NAME, HOUR, MINUTE, SOUND, REPEATING] = response.formValues
                        const hour = HOUR < 10 ? `0${HOUR}` : HOUR
                        const minute = MINUTE < 10 ? `0${MINUTE}` : MINUTE
                        const soundName = alarmSoundTexts[SOUND]
                        const soundValue = alarmSoundValues[SOUND]
                        if (!NAME) return alarms()
                        Clock.addAlarm(userId, NAME, [hour, minute], [soundName, soundValue], REPEATING, true, false)
                        alarms()
                    })
                }

                function manageAlarm(alarm, order) {
                    const { name, time, isRepeating, isActive, sound } = alarm
                    const { soundName, soundValue } = sound
                    const repeatingText = isRepeating ? "Yes" : "No"
                    const onOrOffText1 = isActive ? "Turn Off" : "Turn On"
                    const onOrOffText2 = isActive ? "§aActive" : "§cNot Active"
                    const onOrOffTexture = isActive ? "off" : "on"
                    const switchStatus = isActive ? false : true
                    const form = new ActionFormData()
                        .title("Alarm")
                        .body(`§7Name: §r${name}\n§7Time: §r${time}\n§7Sound: §r${soundName}\n§7Repeat: §r${repeatingText}\n§7Status: ${onOrOffText2}`)
                        .button(onOrOffText1, `textures/coptaine_smartphone/icons/${onOrOffTexture}`)
                        .button("Edit", "textures/coptaine_smartphone/icons/edit")
                        .button("Delete", "textures/coptaine_smartphone/icons/trash")
                        .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                        .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                    form.show(player).then((response) => {
                        if (response.canceled) return
                        if (response.selection === 0) {
                            const hour = Number(time.split(":")[0])
                            const minute = Number(time.split(":")[1])
                            Clock.editAlarm(userId, order, name, [hour, minute], [soundName, soundValue], isRepeating, switchStatus, false)
                            alarms()
                        }
                        if (response.selection === 1) editAlarm()
                        if (response.selection === 2) removeAlarm()
                        if (response.selection === 3) alarms()
                        if (response.selection === 4) smartphone(player)
                    })

                    function editAlarm() {
                        const hour = Number(time.split(":")[0])
                        const minute = Number(time.split(":")[1])
                        const form = new ModalFormData()
                            .title(name)
                            .textField("Name:", name, name)
                            .slider("Hour", 0, 23, 1, hour)
                            .slider("Minute", 0, 59, 1, minute)
                            .dropdown("Sound:", alarmSoundTexts)
                            .toggle("Repeat", isRepeating)
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            const [NAME, HOUR, MINUTE, SOUND, REPEATING] = response.formValues
                            const hour = HOUR < 10 ? `0${HOUR}` : HOUR
                            const minute = MINUTE < 10 ? `0${MINUTE}` : MINUTE
                            const soundName = alarmSoundTexts[SOUND]
                            const soundValue = alarmSoundValues[SOUND]
                            Clock.editAlarm(userId, order, NAME, [hour, minute], [soundName, soundValue], REPEATING, isActive, false)
                            alarms()
                        })
                    }

                    function removeAlarm() {
                        const form = new ActionFormData()
                            .title(`${name}`)
                            .body("§7Do you want to delete this alarm?\n\n")
                            .button("Delete")
                            .button("Cancel")
                        form.show(player).then((response) => {
                            if (response.canceled) return
                            if (response.selection === 0) {
                                Clock.removeAlarm(userId, order)
                                alarms()
                            }
                            if (response.selection === 1) manageAlarm(alarm)
                        })
                    }
                }
            }

            function changeTime(getTime) {
                if (balance < 10000) return timeChangeFailed()
                dimension.runCommand(`time set ${getTime}`)
                timeChangedSuccessful(getTime)
            }

            function timeChangedSuccessful(getTime) {
                const timeString = getTime[0].toUpperCase() + getTime.slice(1)
                const form = new ActionFormData()
                    .title("Clock")
                    .body(`§aSuccessfully changed time to ${timeString}!\n\n`)
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.selection === 0) clock()
                    if (response.selection === 1) smartphone(player)
                })
            }

            function timeChangeFailed() {
                const form = new ActionFormData()
                    .title("Clock")
                    .body(`§cFailed to changed time: Not enough money!\n§7Money: §g$${balance}\n§7You need §g$10000\n\n`)
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.selection === 0) clock()
                    if (response.selection === 1) smartphone(player)
                })
            }
        }

        function statistics() {
            const timePlayed = player.getDynamicProperty("stats:time_played") ?? 0
            const timeSinceLastDeath = player.getDynamicProperty("stats:time_since_last_death") ?? 0
            const minedBlocks = player.getDynamicProperty("stats:mined_blocks") ?? 0
            const placedBlocks = player.getDynamicProperty("stats:placed_blocks") ?? 0
            const foodsEaten = player.getDynamicProperty("stats:foods_eaten") ?? 0
            const deaths = player.getDynamicProperty("stats:deaths") ?? 0
            const mobsKilled = player.getDynamicProperty("stats:mobs_killed") ?? 0
            const playersKilled = player.getDynamicProperty("stats:players_killed") ?? 0
            const dragonKilled = player.getDynamicProperty("stats:dragon_killed") ?? 0
            const witherKilled = player.getDynamicProperty("stats:wither_killed") ?? 0
            const distanceTraveled = player.getDynamicProperty("stats:distance_traveled") ?? 0
            const message = [
                `§7Time Played: §r${(timePlayed / 3600).toFixed(2)} h`,
                `§7Time Since Last Death: §r${(timeSinceLastDeath / 3600).toFixed(2)} h`,
                `§7Distance Traveled: §r${Math.floor(distanceTraveled)} blocks`,
                `§7Blocks Mined: §r${minedBlocks}`,
                `§7Blocks Placed: §r${placedBlocks}`,
                `§7Foods Eaten: §r${foodsEaten}`,
                `§7Deaths: §r${deaths}`,
                `§7Mobs Killed: §r${mobsKilled}`,
                `§7Players Killed: §r${playersKilled}`,
                `§7Ender Dragon Killed: §rx${dragonKilled}`,
                `§7Wither Killed: §rx${witherKilled}`
            ]
            const newMessage = message.join("\n")
            const body = player.getDynamicProperty("stats") ? `${newMessage}` : "§7Status: §cNot Active\n§7§oTIP: Activate in Settings\n\n"
            const form = new ActionFormData()
                .title("Statistics")
                .body(`${body}`)
                .button("§q Refresh", "textures/coptaine_smartphone/icons/refresh")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) statistics()
                if (response.selection === 1) apps()
                if (response.selection === 2) smartphone(player)
            })
        }

        function settings() {
            const form = new ActionFormData()
                .title("Settings")
                .button("Home Widgets", "textures/coptaine_smartphone/icons/widgets")
                .button("Notifications", "textures/coptaine_smartphone/icons/notifications")
                .button("Statistics", "textures/coptaine_smartphone/icons/stats")
                .button("About", "textures/coptaine_smartphone/icons/about")
                .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                .button("§1 Home", "textures/coptaine_smartphone/icons/home")
            form.show(player).then((response) => {
                if (response.canceled) return
                if (response.selection === 0) widgets()
                if (response.selection === 1) notifications()
                if (response.selection === 2) stats()
                if (response.selection === 3) about()
                if (response.selection === 4) apps()
                if (response.selection === 5) smartphone(player)
            })

            function widgets() {
                const location = player.getDynamicProperty("widget:location") ?? false
                const dimension = player.getDynamicProperty("widget:dimension") ?? false
                const biome = player.getDynamicProperty("widget:biome") ?? false
                const light = player.getDynamicProperty("widget:light") ?? false
                const moonPhase = player.getDynamicProperty("widget:moon_phase") ?? false
                const weather = player.getDynamicProperty("widget:weather") ?? false
                const spawnpoint = player.getDynamicProperty("widget:spawnpoint") ?? false
                const form = new ModalFormData()
                    .title("Settings")
                    .toggle("Location", location)
                    .toggle("Dimension", dimension)
                    .toggle("Biome", biome)
                    .toggle("Light Level", light)
                    .toggle("Moon Phase", moonPhase)
                    .toggle("Weather", weather)
                    .toggle("Spawnpoint", spawnpoint)
                form.show(player).then((response) => {
                    if (response.canceled) return
                    const [LOCATION, DIMENSION, BIOME, LIGHT, MOONPHASE, WEATHER, SPAWNPOINT] = response.formValues
                    player.setDynamicProperty("widget:location", LOCATION)
                    player.setDynamicProperty("widget:dimension", DIMENSION)
                    player.setDynamicProperty("widget:biome", BIOME)
                    player.setDynamicProperty("widget:light", LIGHT)
                    player.setDynamicProperty("widget:moon_phase", MOONPHASE)
                    player.setDynamicProperty("widget:weather", WEATHER)
                    player.setDynamicProperty("widget:spawnpoint", SPAWNPOINT)
                    settings()
                })
            }

            function notifications() {
                const message = player.getDynamicProperty("notification:messages") ?? true
                const store = player.getDynamicProperty("notification:store") ?? true
                const banking = player.getDynamicProperty("notification:bank") ?? true
                const form = new ModalFormData()
                    .title("Notifications")
                    .toggle("Message", message)
                    .toggle("Store", store)
                    .toggle("Bank Transactions", banking)
                form.show(player).then((response) => {
                    if (response.canceled) return
                    const [MESSAGE, STORE, BANKING] = response.formValues
                    player.setDynamicProperty("notification:message", MESSAGE)
                    player.setDynamicProperty("notification:store", STORE)
                    player.setDynamicProperty("notification:bank", BANKING)
                    settings()
                })
            }

            function stats() {
                const statsOnOrOff = player.getDynamicProperty("stats")
                const body = statsOnOrOff ? "Deactivate" : "Activate"
                const state = statsOnOrOff ? "§aActive" : "§cNot Active"
                const texture = statsOnOrOff ? "textures/coptaine_smartphone/icons/off" : "textures/coptaine_smartphone/icons/on"
                const form = new ActionFormData()
                    .title("Settings")
                    .body(`§7Status: ${state}\n\n§o§e[WARNING] Turning this on may cause lag!`)
                    .button(`${body}`, texture)
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.selection === 0) {
                        if (statsOnOrOff) {
                            player.setDynamicProperty("stats", false)
                        } else {
                            player.setDynamicProperty("stats", true)
                        }
                        stats()
                    }
                    if (response.selection === 1) settings()
                    if (response.selection === 2) smartphone(player)
                })
            }

            function about() {
                const registeredDate = player.getDynamicProperty("stats:date_registered")
                const form = new ActionFormData()
                    .title("About")
                    .body(`§7Owner: §r${userName}\n§7Registered ${registeredDate}\n§7Coptaine OS`)
                    .button("§4 Back", "textures/coptaine_smartphone/icons/back")
                    .button("§1 Home", "textures/coptaine_smartphone/icons/home")
                form.show(player).then((response) => {
                    if (response.canceled) return
                    if (response.selection === 0) settings()
                    if (response.selection === 1) smartphone(player)
                })
            }
        }
    }
}



PostWorld.playerSpawn.subscribe((e) => {
    const { player, initialSpawn } = e
    if (initialSpawn) {
        player.setDynamicProperty("music", undefined) // Removes Now Playing in music
    }
    if (player.hasTag("viewing_camera")) {
        player.removeTag("viewing_camera")
        player.camera.clear()
    }
})



PostWorld.entityDie.subscribe((e) => {
    const { damageSource, deadEntity } = e
    const killer = damageSource.damagingEntity

    if (deadEntity.typeId === "minecraft:player" && deadEntity.getDynamicProperty("stats")) {
        const deaths = deadEntity.getDynamicProperty("stats:deaths") ?? 0
        deadEntity.setDynamicProperty("stats:deaths", deaths + 1)
        deadEntity.setDynamicProperty("stats:time_since_last_death", 0)

        if (killer?.typeId === "minecraft:player" && killer.getDynamicProperty("stats")) {
            const playersKilled = killer.getDynamicProperty("stats:players_killed") ?? 0
            killer.setDynamicProperty("stats:players_killed", playersKilled + 1)
        }
    } else {
        try {
            if (!killer.getDynamicProperty("stats")) return
            const mobsKilled = killer.getDynamicProperty("stats:mobs_killed") ?? 0
            killer?.setDynamicProperty("stats:mobs_killed", mobsKilled + 1)

            if (deadEntity.typeId === "minecraft:ender_dragon") {
                const dragonKilled = killer.getDynamicProperty("stats:dragon_killed") ?? 0
                killer?.setDynamicProperty("stats:dragon_killed", dragonKilled + 1)
            }
            if (deadEntity.typeId === "minecraft:wither") {
                const witherKilled = killer.getDynamicProperty("stats:wither_killed") ?? 0
                killer?.setDynamicProperty("stats:wither_killed", witherKilled + 1)
            }
        } catch { }
    }
})



PostWorld.itemUse.subscribe((e) => {
    const { itemStack: item, source: player } = e
    const { dimension } = player
    if (item.typeId === "coptaine:smartphone") {
        if (!player.hasTag("viewing_camera")) {
            weatherAtPosition(player)
            system.runTimeout(() => {
                smartphone(player)
                player.playSound("random.pop", { volume: 1.0, pitch: 1.5 })
            }, 5)
        } else {
            if (player.isSneaking) {
                const tpLocation = player.getDynamicProperty("camera_teleport")
                player.removeTag("viewing_camera")
                player.camera.clear()
                if (Bank.balance(player.id) < 30) return player.sendMessage("§p[CAMERA PRO] §cTeleport Failed: Not enough money!")
                player.camera.fade({ fadeColor: { red: 0, green: 0, blue: 0 }, fadeTime: { fadeInTime: 0, fadeOutTime: 0.2, holdTime: 1.2 } })
                player.playAnimation("animation.player.camera_tp")
                dimension.spawnParticle("coptaine:teleport", player.location)
                player.teleport(tpLocation)
                player.playSound("coptaine_us.camera.teleport1")
                Bank.transfer(30, player.id, 0, player.name, "§nTeleport Service")
                system.runTimeout(() => {
                    player.playSound("coptaine_us.camera.teleport2")
                    dimension.spawnParticle("coptaine:teleport", player.location)
                }, 20)
            } else {
                player.removeTag("viewing_camera")
                player.camera.clear()
                smartphone(player, "camera")
            }
        }

    }
    if (item.typeId === "coptaine:holo_billboard") {
        const location = getFaceHitBlock(player)?.location
        const rotation = getFaceHitBlock(player)?.orientation
        if (!location || !rotation) return
        const billboard = dimension.spawnEntity("coptaine:holo_billboard", location)
        billboard.triggerEvent(rotation)
        deductItemInHand(player)
    }

    if (item.typeId === "coptaine:camera") {
        const x = Math.floor(-player.getRotation().x)
        const y = Math.floor(player.getRotation().y + 180)
        const tpLocation = player.location
        const location = getFaceHitBlock(player)?.location
        const rotation = getFaceHitBlock(player)?.orientation
        const myCameras = Camera.getMyCameras(player.id).length
        if (!location) return
        if (!PlayersRegistry.checkStatus(player.id)) return player.sendMessage("§cActivate your smartphone first!")
        if (myCameras === 5) return player.sendMessage("§p[CAMERA] §cMaximum camera limit per player is reached (5/5)!")
        const camera = dimension.spawnEntity("coptaine:camera", location)
        Camera.addCamera(player.id, camera.id, location, { x: x, y: y }, dimension, tpLocation)
        camera.triggerEvent(rotation)
        deductItemInHand(player)
    }
})



PostWorld.entityHitEntity.subscribe((e) => {
    const { hitEntity, damagingEntity } = e
    if (!damagingEntity.typeId === "minecraft:player" || hitEntity.typeId !== "coptaine:holo_billboard") return

    const item = damagingEntity.getComponent("equippable")?.getEquipment("Mainhand")
    if (damagingEntity.matches({ gameMode: "adventure" })) return

    if (item?.typeId !== "coptaine:printed_note") return hitEntity.triggerEvent("e:remove")

    if (hitEntity.getProperty("coptaine:power") === "on") return

    const messageArr = (item.nameTag).split("\n")
    const message = messageArr.join("\n")
    hitEntity.triggerEvent("e:turn_on")
    hitEntity.nameTag = message
    deductItemInHand(damagingEntity)
})



PostWorld.itemCompleteUse.subscribe((e) => {
    const { itemStack: item, source: player } = e
    if (!player.getDynamicProperty("stats")) return
    const isFood = item.getComponent("minecraft:food")?.typeId
    const foodEaten = player.getDynamicProperty("stats:foods_eaten") ?? 0
    if (isFood || vanillaFoods.includes(item.typeId)) return player.setDynamicProperty("stats:foods_eaten", foodEaten + 1)
})



PostWorld.playerBreakBlock.subscribe((e) => {
    const { player } = e
    if (!player.getDynamicProperty("stats")) return
    const blocks = player.getDynamicProperty("stats:mined_blocks") ?? 0
    player.setDynamicProperty("stats:mined_blocks", blocks + 1)
})



PostWorld.playerPlaceBlock.subscribe((e) => {
    const { player } = e
    if (!player.getDynamicProperty("stats")) return
    const blocks = player.getDynamicProperty("stats:placed_blocks") ?? 0
    player.setDynamicProperty("stats:placed_blocks", blocks + 1)
})



system.runInterval(() => {
    for (const player of world.getPlayers()) {
        // Alarm
        const alarms = Clock.getMyAlarms(player.id)
        const container = getInventory(player).container
        let hasSmartphone
        for (let i = 0; i < container.size; i++) {
            if (container.getItem(i)?.typeId === "coptaine:smartphone") {
                hasSmartphone = true
                break
            }
        }
        for (const alarm of alarms) {
            const time = getWorldDayTime().split("/ ")[1]
            if (alarm.isActive && alarm.time == time && hasSmartphone) {
                player.sendMessage(`§n[ALARM] §r${alarm.name} §7[${alarm.time}]§r!`)
                player.playSound(alarm.sound.value)
                alarm.isActive = false
                alarm.isRinging = true
                Clock.serializeData(player.id, alarms)
                system.runTimeout(async () => {
                    if (alarm.isRepeating) {
                        alarm.isActive = true
                        Clock.serializeData(player.id, alarms)
                    }
                }, 20)
            }
        }
        // Statistics
        if (player.getDynamicProperty("stats")) {
            const velocity = player.getVelocity()
            const blocks = player.getDynamicProperty("stats:distance_traveled") ?? 0
            const speed = Math.hypot(velocity.x, velocity.y, velocity.z)
            if (speed) player.setDynamicProperty("stats:distance_traveled", (Number(blocks) + speed).toFixed(2))
            getTimePlayed(player)
            getTimeSinceLastDeath(player)
        }
    }
}, 1)



system.afterEvents.scriptEventReceive.subscribe((event) => {
    const { sourceEntity: entity, message } = event
    const players = world.getPlayers()

    if (event.id === "coptaine:reset_all") {
        for (const player of players) {
            player.clearDynamicProperties()
        }
        world.clearDynamicProperties()
        entity.sendMessage("§acleared!")
    }

    if (event.id === "coptaine:give_money") {
        world.setDynamicProperty(`bank:${entity.id}`, Number(message))
        entity.sendMessage(`§a${message}!`)
    }

    if (entity.typeId === "coptaine:sensor") {
        const { x, z } = entity.location
        const callersHeight = Number(entity.nameTag)
        const playersForStats = entity.dimension.getPlayers({
            location: { x: x, y: callersHeight, z: z },
            maxDistance: 3
        })
        if (event.id === "coptaine:weather") {
            for (const player of playersForStats) {
                player.setDynamicProperty("coptaine:weather", message)
            }
        }

        if (event.id === "coptaine:light") {
            for (const player of playersForStats) {
                player.setDynamicProperty("coptaine:light", message)
            }
        }

        if (event.id === "coptaine:biome") {
            for (const player of playersForStats) {
                player.setDynamicProperty("coptaine:biome", message)
            }
        }

        if (event.id === "coptaine:temperature") {
            for (const player of playersForStats) {
                player.setDynamicProperty("coptaine:temperature", message)
            }
        }
    }
})